
#include "VisionExpert.h"
#include <vector>
#include <boost/program_options.hpp>
//#include "MatrixFunctions.h"
#include "msg.h"
//#include <WheelchairLogger.h>
//#include <WheelChairUtil.h>

#include "jpegmem.h"

#define VISIONDEBUG

using namespace std;

bool VisionExpert::isReady(){
	return false;// ( spykee->getState() & Spykee::CameraOn ) == Spykee::CameraOn;
}

VisionExpert::VisionExpert(ModuleDispatch* kernel,int period, Spykee *s, bool simFileCamera,string prefix,string postfix,string format) : StringModuleMember(kernel,period,"vision"){
	spykee = s;
	
	/* m_image è la variabile dove, a ogni ciclo del codice, viene memorizzata l'immagine ricevuta da spykyy. E' un immmagine di 320x240px, con 3 canali
	   del colore (RGB, a 8 bit di profondità.
	   A ogni iterazione l'immagine precedente viene sovrascritta */
	m_image = cvCreateImage( cvSize( 320, 240 ), 8, 3 );
	
	/* Buffer di byte utilizzato per contenere l'immagine convertita da jpeg a bmp */	
	m_pBmpBuffer = new unsigned char[ 320 * 240 * 3 ];

	/* Contiene il backproject dell'immagine. Deve avere la stessa dimensione dell'immagine originale, ma è in scala di grigi*/
	m_backProject = cvCreateImage( cvSize( 320, 240 ), 8, 1 );
	
	/* Memorizza la maschera in scala di grigi ottenuta effettuando la sogliatura dell'immagine originale*/
	m_mask = cvCreateImage( cvSize( 320, 240 ), 8, 1 );
	
	/* Immagine originale converitita in HSV */
	m_hsv = cvCreateImage( cvSize( 320, 240 ), 8, 3 );
	
	/* Canale HUE che contiene l'immagine convertita */
	m_hue = cvCreateImage( cvSize( 320, 240 ), 8, 1 );

	/* Utilizzata per il debug */
	m_debug = cvCreateImage( cvSize(320,240), 8, 3 );
	memset( m_lastCount, 0, 10 );

	/* Buffer per memorizzare una colonna dell'immagine */
	m_colArr = cvCreateMat( 240, 1, CV_8UC1 );
	
	/* Buffer per memorizzare il numero di pixel su ogni colonna */
	m_posArr = cvCreateMat( 1, 320, CV_8UC1 );

	cvNamedWindow( "TEST", 1 );
	cvNamedWindow( "back", 1 );

#ifdef SAVEVIDEO
	m_imgWriter = cvCreateVideoWriter( "./video.avi", CV_FOURCC( 'd', 'i', 'v', 'x' ), 25, cvSize( 320, 240 ), 1 );
	m_bpWriter = cvCreateVideoWriter( "./backproject.avi", CV_FOURCC( 'd', 'i', 'v', 'x' ), 25, cvSize( 320, 240 ), 1 );
#endif

	/* Istogramma del colore dell'immagine di riferimento del marker */
	m_hist = CreateMatchingHist( "./pink1.jpg" );

	printf( "\nVisionExpert (Spykee edition) build" );
	fflush(stdout);
}

VisionExpert::~VisionExpert (){
	cvReleaseImage( &m_image );
	cvReleaseImage( &m_hsv );
	cvReleaseImage( &m_hue );
	cvReleaseImage( &m_backProject );
	cvReleaseImage( &m_mask );

	cvReleaseImage( &m_debug );

	cvReleaseMat( &m_colArr );
	cvReleaseMat( &m_posArr );

	cvReleaseHist( &m_hist );

	delete [] m_pBmpBuffer;

#ifdef SAVEVIDEO
	cvReleaseVideoWriter( &m_imgWriter );
	cvReleaseVideoWriter( &m_bpWriter );
#endif

	cvDestroyAllWindows();
}

void VisionExpert::RunInit(){
//	cvNamedWindow( "img", 1 );
//	cvWaitKey(100);
//	cvShowImage( "img", image ); 	

	search_area = cvRect( 0, 0, 320, 240 );
}

void VisionExpert::RunClose(){
//	WheelchairLogger::writeLog("Vision Expert run close");
//	if(camera!=NULL)camera->Stop();
}

void VisionExpert::RunDuty(){
	struct timeval 	tstamp,tdiff;
	struct timezone tz;
	char s[256];
	
	int flag=0,flag2=0;
	int notnull;
	int posvalid;
	float posx,posy,posa;
	weightfloat nposx,nposy,nposa;
	weightfloat newdata;
	
	float sx,sy,sa;
	float radiuscorrection;
	float anglecorrection;

	if((spykee->getState() & Spykee::CameraOn) == 0) {
//		MrtLogger::writeLog("Vision Expert :: no camera");	
		printf("no camera!\n");
		return;
	}
	
	unsigned char *tmp = NULL;
	int imgLen = 0;
	unsigned char *jpeg = Spykee::pharseMessage(NULL, &imgLen, &tmp, Spykee::takeCameraPictureWithLock);

	if (imgLen <= 0)
	{
		Spykee::pharseMessage(NULL, NULL, NULL, Spykee::unlockPicture);
//		MrtLogger::writeLog("Vision Expert :: immagine di lunghezza 0");
		return;
	}

	CvConnectedComp track_comp;
	CvBox2D track_box;

	unsigned char *pBmp = jpeg2bmp( jpeg, imgLen );
	
	if ( pBmp != NULL ){
		memcpy( m_image->imageData, pBmp, 320 * 240 * 3 );

		cvCvtColor( m_image, m_image, CV_BGR2RGB );
		cvCvtColor( m_image, m_hsv, CV_RGB2HSV );
		cvInRangeS( m_hsv, cvScalar( 128, 30, 0, 0 ), cvScalar( 135, 255, 255, 0 ), m_mask );

		cvSplit( m_hsv, m_hue, NULL, NULL, NULL );
	
		cvCalcBackProject( &m_hue, m_backProject, m_hist );
		cvAnd( m_backProject, m_mask, m_backProject );

	 	cvSmooth( m_backProject, m_backProject );
		cvThreshold( m_backProject, m_backProject, 50,  255, CV_THRESH_BINARY );

		cvCamShift( m_backProject, search_area,	cvTermCriteria( CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 1, 1 ), &track_comp, &track_box );

		search_area = track_comp.rect;
		track_box.angle = -track_box.angle;

		int current = cvCountNonZero( m_backProject );
#ifdef VISIONDEBUG
	//	printf( "current: %d - %f%%\n", current, ((double)current / ( 320*240)) * 100 );fflush(stdout);
#endif

		cvEllipseBox( m_image, track_box, CV_RGB(255,0,0), 3, 1, 0 );
		cvRectangle( m_image, cvPoint( search_area.x, search_area.y ), cvPoint( search_area.x + search_area.width, search_area.y + search_area.height ), CV_RGB( 0, 255, 0 ) );

		cvShowImage( "TEST", m_image );
		cvShowImage( "back", m_backProject );

		if ( ( cvWaitKey(10) & 8 ) != 27 ){
			cvSaveImage( "./image.jpg", m_image );
		}

	#ifdef SAVEVIDEO
		cvWriteFrame( m_imgWriter, m_image );
		cvWriteFrame( m_bpWriter, m_backProject );
	#endif

		if ( verifyPresence( current ) ){
			for ( int i = 0; i < 320; i++ ){
				cvGetCol( m_backProject, m_colArr, i );
				cvSet1D( m_posArr, i, cvScalar( cvCountNonZero(m_colArr), 0, 0, 0 ) );
			}

			CvPoint max_pos;
			double max_val;

			cvMinMaxLoc( m_posArr, NULL, &max_val, NULL, &max_pos );
			float angle = ( (float)max_pos.x * 46.0 ) / 320.0;

			if ( angle >= 0 || angle <= 46 ){
				printf( "VISION OK: %.3f\n", angle );
				NewMessage( MSG_TO_BRIAN, VISION_DATA );
				AddBrianData( "EnemyDetected", 1, 1 );
				AddBrianData( "EnemyAngle", angle, 1 );
				CloseMessage();
				SendAllStringMessages();
			}
		} else {
	//		printf( "VISION OFF\n" );
			NewMessage( MSG_TO_BRIAN, VISION_DATA );
			AddBrianData( "EnemyDetected", 0, 1 );
			AddBrianData( "EnemyAngle", 0, 1 );
			CloseMessage();
			SendAllStringMessages();

			search_area = cvRect( 0, 0, 320, 240 );
		}
	}

	Spykee::pharseMessage( NULL, NULL, NULL, Spykee::unlockPicture );
}

bool VisionExpert::verifyPresence( int current ){
	int imgSize = 320 * 240;

	if ( current <= 0 )
		return false;

	if ( ( (double)current / imgSize ) > 0.03 )
		return true;

	return false;	
}

unsigned char *VisionExpert::jpeg2bmp( unsigned char *jpeg, int len ){
	jpeg_decompress_struct cinfo;
    jpeg_my_error_mgr jerr;

    cinfo.err = jpeg_std_error( &jerr.pub );
    jpeg_create_decompress( &cinfo );

	jerr.pub.error_exit = my_error_exit;
	jerr.err = false;

	if ( setjmp( jerr.setjmp_buffer ) ){
		jpeg_destroy_decompress(&cinfo);
		return NULL;
	}

    jpeg_memory_src( &cinfo, jpeg, len );

	if ( jerr.err ){
		printf( "ERR2\n" );fflush(stdout);
		jpeg_destroy_decompress( &cinfo );
		return NULL;
	}

    if ( jpeg_read_header( &cinfo, true ) != JPEG_HEADER_OK )
		return NULL;

    unsigned char *pBmp = m_pBmpBuffer;

    jpeg_start_decompress(&cinfo);

    while ( cinfo.output_scanline < cinfo.output_height ){
        pBmp = m_pBmpBuffer + 3 * cinfo.image_width * cinfo.output_scanline;

        jpeg_read_scanlines( &cinfo, &pBmp, 1 );
    }

    jpeg_finish_decompress( &cinfo );
    jpeg_destroy_decompress( &cinfo );

	return m_pBmpBuffer;
}

void VisionExpert::TrackColor( IplImage *image ){

}

void VisionExpert::my_error_exit( j_common_ptr cinfo ){
	struct jpeg_my_error_mgr *err = (jpeg_my_error_mgr*)cinfo->err;
	jpeg_abort(cinfo);
	err->err = true;
//	err->pub.output_message(cinfo);
	longjmp(err->setjmp_buffer, 1 );
}

CvHistogram *VisionExpert::CreateMatchingHist( string file ){
	int hueDim = 32;
	float hueRange[2] = { 0, 180 };

	IplImage *hsv = NULL,
			 *hue = NULL,
			 *mask = NULL;

	CvHistogram *hist = NULL;
	float maxValue = 0;
	int maxIndex = 0;
	printf( "Hist start\n" );

	float *pHueRange = (float*)hueRange;
	//int *pHueDim = (int*)hueDim;

	IplImage *image = cvLoadImage( file.c_str(), 1 );

	if ( image != NULL ){
		hsv = cvCreateImage( cvGetSize( image ), image->depth, image->nChannels );
		hue = cvCreateImage( cvGetSize( image ), 8, 1 );
//		mask = cvCreateImage( cvGetSize(image), 8, 1 );
		hist = cvCreateHist( 1, &hueDim, CV_HIST_ARRAY, &pHueRange, 1 );
	}
	cvCvtColor( image, hsv, CV_RGB2HSV );
	cvSplit( hsv, hue, NULL, NULL, NULL );

	cvCalcHist( &hue, hist, 0 );
	cvGetMinMaxHistValue( hist, 0, &maxValue, 0, &maxIndex );	//valori massimi
	cvConvertScale( hist->bins, hist->bins, maxValue ? 255.0 / maxValue : 0.0, 0 );	//riscalatura

	cvReleaseImage( &image );
	cvReleaseImage( &hsv );
	cvReleaseImage( &hue );
//	cvReleaseImage( &mask );

	return hist;
}

void VisionExpert::DrawHist( CvHistogram *hist, IplImage *image ){
	int hueDim = 32;
	float hueRange[2] = { 0, 180 };

	cvZero( image );

	int binWidth = image->width / hueDim;

	for ( int i = 0; i < hueDim; i++ ){
		float binHeight = cvQueryHistValue_1D( hist, i ) * image->height / 255;
		CvScalar color = hsv2rgb( i * 180.0 / hueDim );

		cvRectangle( image,
					 cvPoint( i * binWidth, image->height ),								//BottomLeft
					 cvPoint( ( i + 1 ) * binWidth, image->height - binHeight ),			//TopRight
					 color, -1, 8, 0 );
		}
}

CvScalar VisionExpert::hsv2rgb( float hue ){
	int rgb[3], p, sector;
	static const int sector_data[][3] = {{0,2,1}, {1,2,0}, {1,0,2}, {2,0,1}, {2,1,0}, {0,1,2}};
	hue *= 0.033333333333333333333333333333333f;
	sector = cvFloor(hue);
	p = cvRound(255*(hue - sector));
	p ^= sector & 1 ? 255 : 0;

	rgb[sector_data[sector][0]] = 255;
	rgb[sector_data[sector][1]] = 0;
	rgb[sector_data[sector][2]] = p;
	
	return cvScalar(rgb[0], rgb[1], rgb[2], 0 );
}

void VisionExpert::AddBrianData(char* name,float value,float rel)
{
	float v;
	float r;
	
	v=value;
	r=rel;
	if(fabs(v)<0.001){ //evito che vengano mandati valori in notazione scientifica
		v=0;
	}
	if(fabs(r)<0.001){
		v=0;
	}
	
	mCurrentMessage<<"<D>"<<name<<" "<<v<<" "<<r<<" </D>\n";
	return;
	
}
